import tkinter as tk
from ..terminal import Terminal


class CommandPrompt(Terminal):
    """Command Prompt - Checks for Command Prompt executable in path and opens that in terminal. 
    Shows Not Available in case variable is not set."""
    
    shell = "C:\\windows\\system32\\cmd.exe"
    name = "Default"
    icon = "cmd"

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

        self.available = self.check_shell()
        if not self.available:
            tk.Label(self, text="Command Prompt not available, report an issue otherwise.").grid()
            self.name = "Not Available"
            self.icon = "error"
            return

        self.start_service()



import os
import tkinter as tk
from threading import Thread
from tkinter import ttk

if os.name == 'nt':
    from winpty import PtyProcess as PTY
else:
    from ptyprocess import PtyProcessUnicode as PTY

from mono.theme import Theme
from mono.utils import Scrollbar

from .ansi import replace_newline, strip_ansi_escape_sequences
from .text import TerminalText


class Terminal(ttk.Frame):
    """Terminal abstract class. All shell types should inherit from this class.
    The inherited class should implement following attributes:
        name (str): Name of the terminal.
        shell (str): command / path to shell executable.
        
    Args:
        master (tk.Tk): Main window.
        cwd (str): Working directory."""
    
    name: str
    shell: str

    def __init__(self, master, cwd=".", theme: Theme=None, standalone=True, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.standalone = standalone
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.alive = False
        self.cwd = cwd
        self.p = None

        if self.standalone:
            self.base = self

            from .styles import Styles
            from .theme import Theme
            self.theme = theme or Theme()
            self.style = Styles(self, self.theme)
        else:
            self.base = master.base
            self.theme = self.base.theme

        self.text = TerminalText(self, relief=tk.FLAT, padx=10, pady=10, font=("Consolas", 11))
        self.text.config(bg=self.theme.terminal[0], fg=self.theme.terminal[1], insertbackground=self.theme.terminal[1])
        self.text.grid(row=0, column=0, sticky=tk.NSEW)
        self.text.bind("<Return>", self.enter)

        self.terminal_scrollbar = Scrollbar(self, style="MonoScrollbar")
        self.terminal_scrollbar.grid(row=0, column=1, sticky='NSW')

        self.text.config(yscrollcommand=self.terminal_scrollbar.set)
        self.terminal_scrollbar.config(command=self.text.yview, orient=tk.VERTICAL)

        self.text.tag_config("prompt", foreground="orange")
        self.text.tag_config("command", foreground="yellow")

        self.bind("<Destroy>", self.stop_service)
    
    def check_shell(self):
        """Check if the shell is available in the system path."""
        
        import shutil
        self.shell = shutil.which(self.shell)
        return self.shell
    
    def start_service(self, *_) -> None:
        """Start the terminal service."""

        self.alive = True
        self.last_command = None
        
        self.p = PTY.spawn([self.shell])
        Thread(target=self._write_loop, daemon=True).start()

    def stop_service(self, *_) -> None:
        """Stop the terminal service."""

        self.alive = False

    def run_command(self, command: str) -> None:
        """Run a command in the terminal.
        TODO: Implement a queue for running multiple commands."""

        self.text.insert("end", command, "command")
        self.enter()

    def enter(self, *_) -> None:
        """Enter key event handler for running commands."""

        command = self.text.get('input', 'end')
        self.last_command = command
        self.text.register_history(command)
        if command.strip():
            self.text.delete('input', 'end')

        self.p.write(command + "\r\n")
        return "break"

    def _write_loop(self) -> None:
        while self.alive:
            if buf := self.p.read():
                p = buf.find('\x1b]0;')
                
                if p != -1:
                    buf = buf[:p]
                buf = [strip_ansi_escape_sequences(i) for i in replace_newline(buf).splitlines()]
                self._insert('\n'.join(buf))

    def _insert(self, output: str, tag='') -> None:
        self.text.insert(tk.END, output, tag)
        #self.terminal.tag_add("prompt", "insert linestart", "insert")
        self.text.see(tk.END)
        self.text.mark_set('input', 'insert')
    
    def _newline(self):
        self._insert('\n')

    def clear(self) -> None:
        """Clear the terminal."""
        
        self.text.clear()

    # TODO: Implement a better way to handle key events.
    def _ctrl_key(self, key: str) -> None:
        if key == 'c':
            self.run_command('\x03')
            
    def __str__(self) -> str:
        return self.name



import tkinter as tk


class TerminalText(tk.Text):
    """Text widget used to display the terminal output and to get the user input.

    Limits the editable area to text after the input mark and prevents deletion
    before the input mark. Also, it keeps a history of previously used commands.

    Args:
        master (tkinter.Tk, optional): The parent widget.
        proxy_enabled (bool, optional): Whether the proxy is enabled. Defaults to True.
    """

    def __init__(self, master=None, proxy_enabled: bool = True, **kw) -> None:
        super().__init__(master, **kw)
        self.master = master

        self.mark_set("input", "insert")
        self.mark_gravity("input", "left")

        self.proxy_enabled = proxy_enabled
        self.config(highlightthickness=0)

        self._history = []
        self._history_level = 0
        self.bind("<Up>", self.history_up)
        self.bind("<Down>", self.history_down)

        self._orig = self._w + "_orig"
        self.tk.call("rename", self._w, self._orig)
        self.tk.createcommand(self._w, self._proxy)

    def history_up(self, *_) -> None:
        """moves up the history and displays it"""

        if not self._history:
            return "break"

        self._history_level = max(self._history_level - 1, 0)
        self.mark_set("insert", "input")
        self.delete("input", "end")
        self.insert("input", self._history[self._history_level])

        return "break"

    def history_down(self, *_) -> None:
        """moves down the history and displays it"""

        if not self._history:
            return "break"

        self._history_level = min(self._history_level + 1, len(self._history) - 1)
        self.mark_set("insert", "input")
        self.delete("input", "end")
        self.insert("input", self._history[self._history_level])

        return "break"

    def register_history(self, command: str) -> None:
        """registers a command in the history"""

        # don't register empty commands or duplicates
        if command.strip() and (
            not self._history or command.strip() != self._history[-1]
        ):
            self._history.append(command.strip())
        self._history_level = len(self._history)

    def clear(self, *_) -> None:
        """clears the text"""

        self.proxy_enabled = False

        lastline = self.get("input linestart", "input")
        self.delete("1.0", "end")
        self.insert("end", lastline)

        self.proxy_enabled = True

    def _proxy(self, *args) -> None:
        """proxy limits the editable area to text after the input mark
        and prevents deletion before the input mark."""

        if not self.proxy_enabled:
            return self.tk.call((self._orig,) + args)

        try:
            largs = list(args)
            if args[0] == "insert":
                if self.compare("insert", "<", "input"):
                    self.mark_set("insert", "end")
            elif args[0] == "delete":
                if self.compare(largs[1], "<", "input"):
                    if len(largs) == 2:
                        return
                    largs[1] = "input"

            result = self.tk.call((self._orig,) + tuple(largs))
            return result
        except:
            # most probably some tkinter-unhandled exception
            pass



import os
import tkinter as tk
from ..terminal import Terminal


class Default(Terminal):
    """Default Terminal - Checks COMSPEC/SHELL environmental variables set in the host machine
    and opens that in terminal. Shows Not Detected in case variable is not set."""

    # get the correct shell command depending on platform
    shell = os.environ.get('COMSPEC') or os.environ.get('SHELL')
    name = icon = os.path.splitext(os.path.basename(shell))[0]

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

        self.available = self.check_shell()
        if not (self.shell and self.available):
            tk.Label(self, text="No shells detected for the host os, report an issue otherwise.").grid()
            self.name = "Not Detected"
            self.icon = "error"
            return

        self.start_service()



class Theme:
    bg = "#181818"
    fg = "#8B949E"
    abg = "#2C2D2D"
    afg = "#CCCCCC"
    border = "#2A2A2A"

    def __init__(self) -> None:
        self.tabbar = self.bg
        self.tab = (self.bg, self.fg, self.abg, self.afg)
        self.tabs = (self.bg, self.fg)
        self.tab_active = (self.abg, self.afg)
        self.terminal = (self.bg, self.fg)
        self.scrollbar = (self.bg, self.abg)



__version__ = "0.35.0"
__version_info__ = tuple(map(int, __version__.split('.')))

import os
import platform
import subprocess
import tkinter as tk

from mono.styles import Styles
from mono.theme import Theme

from .shells import *
from .tabs import Tabs
from .terminal import Terminal


def get_home_directory() -> str:
    if os.name == 'nt':
        return os.path.expandvars("%USERPROFILE%")
    if os.name == 'posix':
        return os.path.expanduser("~")
    return '.'

class Terminals(tk.Frame):
    """Mono's tabbed terminal manager
    
    Args:
        master (tk.Tk): Main window.
        cwd (str): Working directory.
        theme (Theme): Custom theme instance."""
    
    def __init__(self, master, cwd: str=None, theme: Theme=None, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base = self

        self.theme = theme or Theme()
        self.styles = Styles(self, self.theme)

        self.config(bg=self.theme.border)
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid_propagate(False)

        self.cwd = cwd

        self.tabs = Tabs(self)
        self.tabs.grid(row=0, column=1, padx=(1, 0), sticky=tk.NS)

        self.active_terminals = []

    def add_default_terminal(self, *_) -> Default:
        """Add a default terminal to the list. Create a tab for it.
        
        Returns:
            Default: Default terminal instance."""
        
        default_terminal = Default(self, cwd=self.cwd or get_home_directory(), standalone=False)
        self.add_terminal(default_terminal)
        return default_terminal

    def add_terminals(self, terminals) -> None:
        """Append multiple terminals to list. Create tabs for them.

        Args:
            terminals (list): List of Shell types to append."""

        for terminal in terminals:
            self.add_terminal(terminal)

    def add_terminal(self, terminal: Terminal) -> None:
        """Append terminal to list. Create tab for it.

        Args:
            terminal (Terminal): Shell type to append."""

        self.active_terminals.append(terminal)
        self.tabs.add_tab(terminal)
    
    def set_cwd(self, cwd: str) -> None:
        """Set current working directory for all terminals.
        
        Args:
            cwd (str): Directory path."""
        
        self.cwd = cwd

    def open_shell(self, shell: Terminal) -> None:
        """Creates an instance and opens a shell.

        Args:
            shell (Terminal): Shell type to open (not instance)
                use add_terminal() to add existing instance."""
        
        self.add_terminal(shell(self, cwd=self.cwd or get_home_directory(), standalone=False))

    def open_another_terminal(self, cwd: str=None) -> None:
        """Opens another instance of the active terminal.
        
        Args:
            cwd (str): Directory path."""
        
        self.add_terminal(self.active_terminal_type(self, cwd=cwd or self.cwd or get_home_directory(), standalone=False))

    def delete_all_terminals(self, *_) -> None:
        """Permanently delete all terminal instances."""

        for terminal in self.active_terminals:
            terminal.destroy()

        self.tabs.clear_all_tabs()
        self.active_terminals.clear()
        self.refresh()

    def delete_terminal(self, terminal: Terminal) -> None:
        """Permanently delete a terminal instance.
        
        Args:
            terminal (Terminal): Terminal instance to delete."""

        terminal.destroy()
        self.active_terminals.remove(terminal)

    def delete_active_terminal(self, *_) -> None:
        """Permanently delete the active terminal."""
        
        try:
            self.tabs.close_active_tab()
        except IndexError:
            pass

    def set_active_terminal(self, terminal: Terminal) -> None:
        """Switch tabs to the terminal.
        
        Args:
            terminal (Terminal): Terminal instance to switch to."""

        for tab in self.tabs.tabs:
            if tab.terminal == terminal:
                self.tabs.set_active_tab(tab)
    
    def set_active_terminal_by_name(self, name: str) -> None:
        """Switch tabs to the terminal by name.
        
        Args:
            name (str): Name of the terminal to switch to."""

        for tab in self.tabs.tabs:
            if tab.terminal.name == name:
                self.tabs.set_active_tab(tab)
                break

    def clear_terminal(self, *_) -> None:
        """Clear text in the active terminal."""

        if active := self.active_terminal:
            active.clear()
        
    def run_command(self, command: str) -> None:
        """Run a command in the active terminal. If there is no active terminal,
        create a default terminal and run the command.
        
        Args:
            command (str): Command to run."""

        if not self.active_terminal:
            default = self.add_default_terminal()
            default.run_command(command)
            # this won't work, TODO: implement a queue for commands
        else:
            self.active_terminal.run_command(command)
    
    @staticmethod
    def run_in_external_console(self, command: str) -> None:
        """Run a command in an external console.
        
        Args:
            command (str): Command to run."""

        match platform.system():
            case 'Windows':
                subprocess.Popen(['start', 'cmd', '/K', command], shell=True)
            case 'Linux':
                subprocess.Popen(['x-terminal-emulator', '-e', command])
            case 'Darwin':
                subprocess.Popen(['open', '-a', 'Terminal', command])
            case _:
                print("No terminal emulator detected.")

    def open_pwsh(self, *_):
        """Create a Powershell terminal instance and open it"""

        self.open_shell(get_shell_from_name("Powershell"))

    def open_cmd(self, *_):
        """Create a Command Prompt terminal instance and open it"""

        self.open_shell(get_shell_from_name("Command Prompt"))

    def open_bash(self, *_):
        """Create a Bash terminal instance and open it"""

        self.open_shell(get_shell_from_name("Bash"))

    def open_python(self, *_):
        """Create a Python terminal instance and open it"""

        self.open_shell(get_shell_from_name("Python"))

    @property
    def active_terminal_type(self) -> Terminal:
        """Get the type of the active terminal. If there is no active 
        terminal, return Default type."""

        if active := self.active_terminal:
            return type(active)

        return Default

    @property
    def active_terminal(self) -> Terminal:
        """Get the active terminal instance."""

        if not self.tabs.active_tab:
            return

        return self.tabs.active_tab.terminal

    def refresh(self, *_) -> None:
        """Generates <<Empty>> event that can be bound to hide the terminal 
        if there are no active terminals."""

        if not self.active_terminals:
            self.event_generate("<<Empty>>", when="tail")



import tkinter as tk

ICONS = {
    "cmd": "terminal-cmd",
    "bash": "terminal-bash",
    "powershell": "terminal-powershell"
}

class Tab(tk.Frame):
    def __init__(self, master, terminal, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base = master.base
        self.theme = self.base.theme

        self.terminal = terminal
        self.selected = False

        self.bg, self.fg, self.hbg, self.hfg = self.theme.tab
        self.config(bg=self.theme.border)

        self.container = tk.Frame(self, bg=self.bg)
        self.container.pack(side=tk.LEFT, fill=tk.X, expand=True) # for lines bw set pady=(0,1)

        self.name_label = tk.Label(self, text=terminal.name  or terminal.__class__.__name__, padx=5, 
                             font=('Segoe UI', 11), anchor=tk.W, bg=self.bg, fg=self.fg)
        self.name_label.pack(in_=self.container, side=tk.LEFT, expand=True, fill=tk.X)

        self.closebtn = tk.Label(self, text='×', font=('Arial', 15), fg=self.fg, bg=self.bg)
        self.closebtn.bind("<Button-1>", self.close)
        self.closebtn.pack(in_=self.container, padx=5)

        self.bind("<Button-1>", self.select)
        self.name_label.bind("<Button-1>", self.select)

        self.bind("<Enter>", self.on_hover)
        self.bind("<Leave>", self.off_hover)

    def close(self, *_) -> None:
        self.master.close_tab(self)

    def on_hover(self, *_) -> None:
        if not self.selected:
            self.name_label.config(bg=self.hbg)
            self.container.config(bg=self.hbg)
            self.closebtn.config(bg=self.hbg)

            self.hovered = True

    def off_hover(self, *_) -> None:
        if not self.selected:
            self.name_label.config(bg=self.bg)
            self.container.config(bg=self.bg)
            self.closebtn.config(bg=self.bg)
            self.hovered = False

    def deselect(self, *_) -> None:
        if self.selected:
            self.terminal.grid_remove()

            self.name_label.config(bg=self.bg, fg=self.fg)
            self.container.config(bg=self.bg)
            self.closebtn.config(bg=self.bg, activeforeground=self.fg)
            self.selected = False

    def select(self, *_) -> None:
        if not self.selected:
            self.master.set_active_tab(self)
            self.terminal.grid(column=0, row=0, sticky=tk.NSEW)

            self.name_label.config(bg=self.hbg, fg=self.hfg)
            self.container.config(bg=self.hbg)
            self.closebtn.config(bg=self.hbg, activeforeground=self.hfg)
            self.selected = True



import tkinter as tk

from ..terminal import Terminal


class Bash(Terminal):
    """Linux Bash - Checks for bash executable in path and opens that in terminal. 
    Shows Not Available in case variable is not set."""

    shell = "/bin/bash"
    name = "Bash"
    icon = "bash"
    
    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

        self.available = self.check_shell()
        if not self.available:
            tk.Label(self, text="Bash not available, report an issue otherwise.").grid()
            self.name = "Not Available"
            self.icon = "error"
            return

        self.start_service()



from tkinter import ttk

from mono.theme import Theme


class Styles(ttk.Style):
    """Custom ttk styles for Mono terminal."""
    
    def __init__(self, master, theme: Theme, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.theme = master.theme
        
        self.config_scrollbar()

    def config_scrollbar(self) -> None:
        self.element_create("MonoScrollbar.trough", "from", "clam")
        self.element_create("MonoScrollbar.thumb", "from", "clam")

        self.layout("MonoScrollbar", [
            ('MonoScrollbar.trough', {
                'sticky': 'ns',
                'children': [
                    ('MonoScrollbar.thumb', {
                        'unit': '1',
                        'sticky': 'nsew'
                    })
                ]
            })
        ])


        bg, highlight = self.theme.scrollbar
        self.configure("MonoScrollbar", gripcount=0, background=bg, troughcolor=bg, bordercolor=bg, lightcolor=bg, darkcolor=bg, arrowsize=14)
        self.map("MonoScrollbar", background=[('pressed', highlight), ('!disabled', self.theme.border)])



import platform

from .bash import Bash
from .cmd import CommandPrompt
from .default import Default
from .powershell import PowerShell
from .python import Python

SHELLS = {
    "Default": Default,
    "Powershell": PowerShell,
    "Python": Python
}

if platform.system() == "Windows":
    SHELLS["Command Prompt"] = CommandPrompt
elif platform.system() == "Linux":
    SHELLS["Bash"] = Bash

def get_available_shells():
    return SHELLS

def get_shell_from_name(name):
    return SHELLS.get(name, Default)

def register_shell(name, shell):
    SHELLS[name] = shell
    
def is_shell_registered(name):
    return name in SHELLS



import tkinter as tk

from .tab import Tab


class Tabs(tk.Frame):
    def __init__(self, master, width=170, *args, **kwargs) -> None:
        super().__init__(master, width=width, *args, **kwargs)
        self.master = master
        self.base = master.base

        self.pack_propagate(False)
        self.config(bg=self.base.theme.tabbar)

        self.tabs = []
        self.active_tab = None

    def add_tab(self, view) -> None:
        tab = Tab(self, view)
        tab.pack(fill=tk.X)
        self.tabs.append(tab)

        tab.select()

    def set_active_tab(self, selected_tab) -> None:
        self.active_tab = selected_tab
        for tab in self.tabs:
            if tab != selected_tab:
                tab.deselect()

    def clear_all_tabs(self) -> None:
        for tab in self.tabs:
            tab.destroy()

        self.tabs.clear()

    def close_active_tab(self) -> None:
        self.close_tab(self.active_tab)

    def close_tab(self, tab) -> None:
        if not self.tabs:
            return
        
        i = self.tabs.index(tab)
        self.tabs.remove(tab)
        tab.terminal.grid_forget()
        self.master.delete_terminal(tab.terminal)
        tab.destroy()

        if self.tabs:
            if i < len(self.tabs):
                self.tabs[i].select()
            else:
                self.tabs[i-1].select()
        else:
            self.active_tab = None
        self.master.refresh()



import re


SEQ = re.compile(r'\x1b(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
NEWLINE = re.compile(r'\x1b\[\d+\;1H')

def strip_ansi_escape_sequences(string):
    return SEQ.sub('', string)

def replace_newline(string):
    return NEWLINE.sub('\n', string)
